@@ -72,16 +72,15 @@
 NULL
 
 #' @rdname subsetting
-#' @inheritParams base::`[.data.frame`
+#' @inheritParams base::Extract.data.frame
+#' @param name A [name] or a string.
 #' @export
 `$.tbl_df` <- function(x, name) {
-  j <- match(as_string(name), names2(x))
-  if (is.na(j)) {
+  out <- .subset2(x, name)
+  if (is.null(out)) {
     warn(paste0("Unknown or uninitialised column: ", tick(name), "."))
-    NULL
-  } else {
-    .subset2(x, j)
   }
+  out
 }
 
 
@@ -90,7 +89,7 @@ NULL
 `$<-.tbl_df` <- function(x, name, value) {
   tbl_subassign(
     x, i = NULL, as_string(name), list(value),
-    i_arg = NULL, j_arg = NULL, value_arg = substitute(value)
+    i_arg = NULL, j_arg = name, value_arg = substitute(value)
   )
 }
 
@@ -156,16 +155,23 @@ NULL
     }
   }
 
-  if (!is_null(i)) {
+  if (!is.null(i)) {
     i <- vectbl_as_row_location2(i, fast_nrow(x), i_arg, assign = TRUE)
   }
 
   value <- list(value)
 
+  if (is.object(j)) {
+    j <- vectbl_as_col_subscript2(j, j_arg, assign = TRUE)
+  }
+
+  # Side effect: check scalar
+  if (length(j) != 1L || (is.numeric(j) && j < 0) || is.logical(j)) {
+    vectbl_as_col_location2(j, length(x) + 1L, j_arg = j_arg, assign = TRUE)
+  }
+
   j <- vectbl_as_new_col_index(j, x, value, j_arg, value_arg)
 
-  # Side effect: check scalar, allow OOB position
-  vectbl_as_col_location2(j, length(x) + 1L, j_arg = j_arg, assign = TRUE)
   # New columns are added to the end, provide index to avoid matching column
   # names again
   names(value) <- names(j)
@@ -218,7 +224,23 @@ NULL
   }
 
   # From here on, i, j and drop contain correct values:
-  xo <- tbl_subset_col(x, j = j, j_arg)
+  xo <- x
+
+  if (!is.null(j)) {
+    j <- vectbl_as_col_location(j, length(xo), names(xo), j_arg = j_arg, assign = FALSE)
+
+    if (anyNA(j)) {
+      cnd_signal(error_na_column_index(which(is.na(j))))
+    }
+
+    xo <- .subset(x, j)
+
+    if (anyDuplicated(j)) {
+      xo <- set_repaired_names(xo, .name_repair = "minimal")
+    }
+
+    xo <- set_tibble_class(xo, nrow = fast_nrow(x))
+  }
 
   if (!is.null(i)) {
     xo <- tbl_subset_row(xo, i = i, i_arg)
@@ -293,9 +315,9 @@ vectbl_as_row_index <- function(i, x, i_arg, assign = FALSE) {
 }
 
 fix_oob <- function(i, n, warn = TRUE) {
-  if (all(i >= 0, na.rm = TRUE)) {
+  if (any(i > 0, na.rm = TRUE)) {
     fix_oob_positive(i, n, warn)
-  } else if (all(i <= 0, na.rm = TRUE)) {
+  } else if (any(i < 0, na.rm = TRUE)) {
     fix_oob_negative(i, n, warn)
   } else {
     # Will throw error in vec_as_location()
@@ -305,26 +327,22 @@ fix_oob <- function(i, n, warn = TRUE) {
 
 fix_oob_positive <- function(i, n, warn = TRUE) {
   oob <- which(i > n)
-  if (warn) {
-    warn_oob(oob, n)
+  if (warn && length(oob) > 0) {
+    deprecate_soft("3.0.0", "tibble::`[.tbl_df`(i = 'must lie in [0, rows] if positive,')",
+                   details = "Use `NA_integer_` as row index to obtain a row full of `NA` values.",
+                   env = foreign_caller_env())
   }
 
   i[oob] <- NA_integer_
   i
 }
 
-warn_oob <- function(oob, n) {
-  if (has_length(oob)) {
-    deprecate_soft("3.0.0", "tibble::`[.tbl_df`(i = 'must lie in [0, rows] if positive,')",
-      details = "Use `NA` as row index to obtain a row full of `NA` values.",
-      env = foreign_caller_env())
-  }
-}
-
 fix_oob_negative <- function(i, n, warn = TRUE) {
   oob <- (i < -n)
-  if (warn) {
-    warn_oob_negative(which(oob), n)
+  if (warn && any(oob, na.rm = TRUE)) {
+    deprecate_soft("3.0.0", "tibble::`[.tbl_df`(i = 'must lie in [-rows, 0] if negative,')",
+                   details = "Use `NA_integer_` as row index to obtain a row full of `NA` values.",
+                   env = foreign_caller_env())
   }
 
   i <- i[!oob]
@@ -332,20 +350,12 @@ fix_oob_negative <- function(i, n, warn = TRUE) {
   i
 }
 
-warn_oob_negative <- function(oob, n) {
-  if (has_length(oob)) {
-    deprecate_soft("3.0.0", "tibble::`[.tbl_df`(i = 'must lie in [-rows, 0] if negative,')",
-      details = "Use `NA` as row index to obtain a row full of `NA` values.",
-      env = foreign_caller_env())
-  }
-}
-
 fix_oob_invalid <- function(i, is_na_orig) {
   oob <- which(is.na(i) & !is_na_orig)
 
   if (has_length(oob)) {
     deprecate_soft("3.0.0", "tibble::`[.tbl_df`(i = 'must use valid row names')",
-      details = "Use `NA` as row index to obtain a row full of `NA` values.",
+      details = "Use `NA_integer_` as row index to obtain a row full of `NA` values.",
       env = foreign_caller_env())
 
     i[oob] <- NA_integer_
@@ -353,16 +363,6 @@ fix_oob_invalid <- function(i, is_na_orig) {
   i
 }
 
-vectbl_as_col_index <- function(j, x, j_arg, assign = FALSE) {
-  stopifnot(!is.null(j))
-
-  if (vec_is(j) && anyNA(j)) {
-    cnd_signal(error_na_column_index(which(is.na(j))))
-  }
-
-  vectbl_as_col_location(j, length(x), names(x), j_arg = j_arg, assign = assign)
-}
-
 tbl_subset2 <- function(x, j, j_arg) {
   if (is.matrix(j)) {
     deprecate_soft("3.0.0", "tibble::`[[.tbl_df`(j = 'can\\'t be a matrix",
@@ -370,48 +370,55 @@ tbl_subset2 <- function(x, j, j_arg) {
       env = foreign_caller_env())
 
     return(as.matrix(x)[[j]])
-  } else if (has_length(j, 2) && is.numeric(j)) {
-    deprecate_soft("3.0.0", "tibble::`[[.tbl_df`(j = 'can\\'t be a vector of length 2')",
-      details = "Recursive subsetting is deprecated for tibbles.",
-      env = foreign_caller_env())
-
-    return(.subset2(x, j))
-  } else if (is.character(j)) {
-    # Side effect: check that j is a scalar and not NA, allow invalid column names
-    vectbl_as_col_location2(j, length(x) + 1L, c(names(x), j), j_arg = j_arg)
-
-    # Don't warn when accessing invalid column names
-    return(.subset2(x, j))
   }
 
-  j <- vectbl_as_col_location2(j, length(x), names(x), j_arg = j_arg)
+  if (is.object(j)) {
+    j <- vectbl_as_col_subscript2(j, j_arg)
+  }
+
+  if (is.numeric(j)) {
+    if (length(j) == 1L) {
+      if (is.na(j) || j < 1 || j > length(x) || (is.double(j) && j != trunc(j))) {
+        # Side effect: throw error for invalid j
+        vectbl_as_col_location2(j, length(x), j_arg = j_arg)
+      }
+    } else if (length(j) == 2L) {
+      deprecate_soft("3.0.0", "tibble::`[[.tbl_df`(j = 'can\\'t be a vector of length 2')",
+        details = "Recursive subsetting is deprecated for tibbles.",
+        env = foreign_caller_env())
+    } else {
+      # Side effect: throw error for invalid j
+      vectbl_as_col_location2(j, length(x), j_arg = j_arg)
+    }
+  } else if (is.symbol(j)) {
+    # FIXME: Only relevant for R < 3.4
+    j <- as.character(j)
+  } else if (is.logical(j) || length(j) != 1L || !is_bare_atomic(j) || is.na(j)) {
+    # Side effect: throw error for invalid j
+    vectbl_as_col_location2(j, length(x), names(x), j_arg = j_arg)
+  }
+
   .subset2(x, j)
 }
 
-tbl_subset_col <- function(x, j, j_arg) {
-  if (is_null(j)) return(x)
-  j <- vectbl_as_col_index(j, x, j_arg = j_arg)
-  xo <- .subset(x, j)
-  xo <- set_repaired_names(xo, .name_repair = "minimal")
-  set_tibble_class(xo, nrow = fast_nrow(x))
-}
-
 tbl_subset_row <- function(x, i, i_arg) {
-  if (is_null(i)) return(x)
+  if (is.null(i)) return(x)
   i <- vectbl_as_row_index(i, x, i_arg)
   xo <- lapply(unclass(x), vec_slice, i = i)
   set_tibble_class(xo, nrow = length(i))
 }
 
 tbl_subassign <- function(x, i, j, value, i_arg, j_arg, value_arg) {
-  i <- vectbl_as_new_row_index(i, x, i_arg)
+  if (!is.null(i_arg) && !is.null(i)) {
+    i <- vectbl_as_new_row_index(i, x, i_arg)
+  }
 
   if (is.null(i)) {
     value <- vectbl_wrap_rhs_col(value, value_arg)
 
     if (is.null(j)) {
       j <- seq_along(x)
-    } else {
+    } else if (!is.null(j_arg)) {
       j <- vectbl_as_new_col_index(j, x, value, j_arg, value_arg)
     }
 
@@ -431,8 +438,10 @@ tbl_subassign <- function(x, i, j, value, i_arg, j_arg, value_arg) {
       # Optimization: match only once
       # (Invariant: x[[j]] is equivalent to x[[vec_as_location(j)]],
       # allowed by corollary that only existing columns can be updated)
-      j <- vectbl_as_new_col_index(j, x, value, j_arg, value_arg)
-      new <- which(j > ncol(x))
+      if (!is.null(j_arg)) {
+        j <- vectbl_as_new_col_index(j, x, value, j_arg, value_arg)
+      }
+      new <- which(j > length(x))
       value <- vectbl_recycle_rhs(value, length(i), length(j), i_arg, value_arg)
 
       # Fill up columns if necessary
@@ -443,7 +452,7 @@ tbl_subassign <- function(x, i, j, value, i_arg, j_arg, value_arg) {
         x <- tbl_subassign_col(x, j[new], init)
       }
 
-      xj <- tbl_subset_col(x, j, j_arg)
+      xj <- .subset(x, j)
       xj <- tbl_subassign_row(xj, i, value, value_arg)
       xo <- tbl_subassign_col(x, j, unclass(xj))
     }
@@ -463,18 +472,22 @@ vectbl_as_new_row_index <- function(i, x, i_arg) {
     nr <- fast_nrow(x)
 
     new <- which(i > nr)
-    i_new <- i[new]
-    i[new] <- NA
+    if (length(new) > 0) {
+      i_new <- i[new]
+      i[new] <- NA
+
+      if (!is_tight_sequence_at_end(i_new, nr)) {
+        cnd_signal(error_new_rows_at_end_only(nr, i_new))
+      }
+    }
 
     # Only update existing, caller knows how to deal with OOB
     i <- vectbl_as_row_location(i, nr, i_arg, assign = TRUE)
 
-    if (!is_tight_sequence_at_end(i_new, nr)) {
-      cnd_signal(error_new_rows_at_end_only(nr, i_new))
-    }
-
     # Restore, caller knows how to deal
-    i[new] <- i_new
+    if (length(new) > 0) {
+      i[new] <- i_new
+    }
     i
   } else if (is_logical(i)) {
     # Don't allow OOB logical
@@ -493,47 +506,56 @@ vectbl_as_new_col_index <- function(j, x, value, j_arg, value_arg) {
   # Values: index
   # Name: column name (for new columns)
 
-  if (vec_is(j) && anyNA(j)) {
-    cnd_signal(error_assign_columns_non_na_only())
-  }
-
   if (is_bare_character(j)) {
+    if (anyNA(j)) {
+      cnd_signal(error_assign_columns_non_na_only())
+    }
+
     out <- match(j, names(x))
     new <- which(is.na(out))
     if (has_length(new)) {
-      out[new] <- seq.int(ncol(x) + 1L, length.out = length(new))
+      out[new] <- seq.int(length(x) + 1L, length.out = length(new))
     }
-    set_names(out, j)
+    j <- set_names(out, j)
   } else if (is_bare_numeric(j)) {
-    if (anyDuplicated(j)) {
-      cnd_signal(error_duplicate_column_subscript_for_assignment(j))
+    if (anyNA(j)) {
+      cnd_signal(error_assign_columns_non_na_only())
     }
 
-    new <- which(j > ncol(x))
-    j_new <- j[new]
-    j[new] <- NA
+    new <- which(j > length(x))
+    if (length(new) > 0) {
+      j_new <- j[new]
+      j[new] <- NA
 
-    j <- vectbl_as_col_location(j, ncol(x), j_arg = j_arg, assign = TRUE)
-
-    if (!is_tight_sequence_at_end(j_new, ncol(x))) {
-      cnd_signal(error_new_columns_at_end_only(ncol(x), j_new))
+      if (!is_tight_sequence_at_end(j_new, length(x))) {
+        cnd_signal(error_new_columns_at_end_only(length(x), j_new))
+      }
     }
 
-    j[new] <- j_new
-
+    j <- vectbl_as_col_location(j, length(x), j_arg = j_arg, assign = TRUE)
     # FIXME: Recycled names are not repaired
     # FIXME: Hard-coded name repair
     names <- vectbl_recycle_rhs_names(names2(value), length(j), value_arg)
-    names[new][names[new] == ""] <- paste0("...", j_new)
 
-    set_names(j, names)
+    if (length(new) > 0) {
+      j[new] <- j_new
+      names[new][names[new] == ""] <- paste0("...", j_new)
+    }
+
+    j <- set_names(j, names)
   } else {
-    j <- vectbl_as_col_index(j, x, j_arg, assign = TRUE)
-    if (anyDuplicated(j)) {
-      cnd_signal(error_duplicate_column_subscript_for_assignment(j))
+    j <- vectbl_as_col_location(j, length(x), names(x), j_arg = j_arg, assign = TRUE)
+
+    if (anyNA(j)) {
+      cnd_signal(error_na_column_index(which(is.na(j))))
     }
-    j
   }
+
+  if (anyDuplicated(j)) {
+    cnd_signal(error_duplicate_column_subscript_for_assignment(j))
+  }
+
+  j
 }
 
 vectbl_as_row_location <- function(i, n, i_arg, assign = FALSE) {
@@ -550,19 +572,23 @@ vectbl_as_row_location <- function(i, n, i_arg, assign = FALSE) {
     i <- i[, 1]
   }
 
-  subclass_row_index_errors(vec_as_location(i, n, arg = as_label(i_arg)), i_arg = i_arg, assign = assign)
+  subclass_row_index_errors(vec_as_location(i, n), i_arg = i_arg, assign = assign)
 }
 
 vectbl_as_row_location2 <- function(i, n, i_arg, assign = FALSE) {
-  subclass_row_index_errors(vec_as_location2(i, n, arg = as_label(i_arg)), i_arg = i_arg, assign = assign)
+  subclass_row_index_errors(vec_as_location2(i, n), i_arg = i_arg, assign = assign)
 }
 
 vectbl_as_col_location <- function(j, n, names = NULL, j_arg, assign = FALSE) {
-  subclass_col_index_errors(vec_as_location(j, n, names, arg = as_label(j_arg)), j_arg = j_arg, assign = assign)
+  subclass_col_index_errors(vec_as_location(j, n, names), j_arg = j_arg, assign = assign)
 }
 
 vectbl_as_col_location2 <- function(j, n, names = NULL, j_arg, assign = FALSE) {
-  subclass_col_index_errors(vec_as_location2(j, n, names, arg = as_label(j_arg)), j_arg = j_arg, assign = assign)
+  subclass_col_index_errors(vec_as_location2(j, n, names), j_arg = j_arg, assign = assign)
+}
+
+vectbl_as_col_subscript2 <- function(j, j_arg, assign = FALSE) {
+  subclass_col_index_errors(vec_as_subscript2(j, logical = "error"), j_arg = j_arg, assign = assign)
 }
 
 is_tight_sequence_at_end <- function(i_new, n) {
@@ -570,7 +596,7 @@ is_tight_sequence_at_end <- function(i_new, n) {
 }
 
 tbl_subassign_col <- function(x, j, value) {
-  is_data <- !map_lgl(value, is_null)
+  is_data <- !vapply(value, is.null, NA)
   nrow <- fast_nrow(x)
 
   x <- unclass(x)
@@ -624,7 +650,7 @@ tbl_subassign_row <- function(x, i, value, value_arg) {
   nrow <- fast_nrow(x)
   x <- unclass(x)
 
-  tryCatch(
+  withCallingHandlers(
     for (j in seq_along(x)) {
       xj <- x[[j]]
       vec_slice(xj, i) <- value[[j]]
@@ -656,12 +682,12 @@ vectbl_strip_names <- function(x) {
 }
 
 vectbl_wrap_rhs_col <- function(value, value_arg) {
-  if (is_null(value)) {
+  if (is.null(value)) {
     return(list(value))
   }
 
   value <- result_vectbl_wrap_rhs(value)
-  if (is_null(value)) {
+  if (is.null(value)) {
     cnd_signal(error_need_rhs_vector_or_null(value_arg))
   }
 
@@ -670,7 +696,7 @@ vectbl_wrap_rhs_col <- function(value, value_arg) {
 
 vectbl_wrap_rhs_row <- function(value, value_arg) {
   value <- result_vectbl_wrap_rhs(value)
-  if (is_null(value)) {
+  if (is.null(value)) {
     cnd_signal(error_need_rhs_vector(value_arg))
   }
 
@@ -680,37 +706,47 @@ vectbl_wrap_rhs_row <- function(value, value_arg) {
 result_vectbl_wrap_rhs <- function(value) {
   if (!vec_is(value)) {
     NULL
+  } else if (is.list(value)) {
+    # Also covers the case of data frames
+    unclass(value)
   } else if (is.array(value)) {
     if (any(dim(value)[-1:-2] != 1)) {
       return(NULL)
     }
     dim(value) <- head(dim(value), 2)
     as.list(as.data.frame(value, stringsAsFactors = FALSE))
-  } else if (is_atomic(value)) {
-    list(value)
   } else {
-    unclass(value)
+    list(value)
   }
 }
 
 vectbl_recycle_rhs <- function(value, nrow, ncol, i_arg, value_arg) {
-  tryCatch(
-    for (j in seq_along(value)) {
-      if (!is.null(value[[j]])) {
-        value[[j]] <- vec_recycle(value[[j]], nrow)
+  if (length(value) > 0L && (nrow != 1L || vec_size(value[[1L]]) != 1L)) {
+    withCallingHandlers(
+      for (j in seq_along(value)) {
+        if (!is.null(value[[j]])) {
+          value[[j]] <- vec_recycle(value[[j]], nrow)
+        }
+      },
+
+      vctrs_error_recycle_incompatible_size = function(cnd) {
+        cnd_signal(error_assign_incompatible_size(nrow, value, j, i_arg, value_arg))
       }
-    },
+    )
+  }
 
-    vctrs_error_recycle_incompatible_size = function(cnd) {
-      cnd_signal(error_assign_incompatible_size(nrow, value, j, i_arg, value_arg))
-    }
-  )
+  if (length(value) != 1L || ncol != 1L) {
+    # Errors have been caught beforehand in vectbl_recycle_rhs_names()
+    value <- vec_recycle(value, ncol)
+  }
 
-  # Errors have been caught beforehand in vectbl_recycle_rhs_names()
-  vec_recycle(value, ncol)
+  value
 }
 
 vectbl_recycle_rhs_names <- function(names, n, value_arg) {
+  if (n == 1L && length(names) == 1L) {
+    return(names)
+  }
   unname(vec_recycle(set_names(names), n, x_arg = as_label(value_arg)))
 }
 
@@ -841,8 +877,8 @@ error_assign_incompatible_type <- function(x, value, j, value_arg, message) {
 # Subclassing errors ------------------------------------------------------
 
 subclass_col_index_errors <- function(expr, j_arg, assign) {
-  tryCatch(
-    force(expr),
+  withCallingHandlers(
+    expr,
     vctrs_error_subscript = function(cnd) {
       cnd$subscript_arg <- j_arg
       cnd$subscript_elt <- "column"
@@ -855,8 +891,8 @@ subclass_col_index_errors <- function(expr, j_arg, assign) {
 }
 
 subclass_row_index_errors <- function(expr, i_arg, assign) {
-  tryCatch(
-    force(expr),
+  withCallingHandlers(
+    expr,
     vctrs_error_subscript = function(cnd) {
       cnd$subscript_arg <- i_arg
       cnd$subscript_elt <- "row"
